#include "stdafx.h"
#include "CommandLine.h"
#include "Utility.h"
#include "GroupDef.h"
#include "SequenceElement.h"
#include "Exception.h"

CommandLineOptions::CommandLineOptions()
	: m_bCrossNumbers(false), m_bOrders(false), m_bMultiplicity(false), m_bDisplayScreen(false),
	m_bUseSequenceFile(true), m_bBlockSearch(false),

	m_eDisplayFormat(DispFormatStandard),
	m_nLengthLowerBound(1), m_nLengthUpperBound(0),
	m_nProgressChangeElement(0), m_nBlockLength(0),

	m_strInFile(""), m_strSequenceFile(""), m_strStatsFile(""),
	m_apGroup(NULL)
{
}

CommandLineOptions::~CommandLineOptions()
{

}

FiniteAbelianGroup& CommandLineOptions::GetGroup()
{
	return *m_apGroup;
}

const FiniteAbelianGroup& CommandLineOptions::GetGroup() const
{
	return *m_apGroup;
}

bool CommandLineOptions::GroupExists() const
{
	return (m_apGroup.get()) ? true : false;
}



void CommandLineOptions::Initialize()
{
	m_ValidOptions.push_back(strInputFileOption);
	m_ValidOptions.push_back(strSequenceFileOption);
	m_ValidOptions.push_back(strStatsFileOption);

	m_ValidOptions.push_back(strCrossNumberOption);
	m_ValidOptions.push_back(strOrderOption);
	m_ValidOptions.push_back(strMultiplicityOption);
	m_ValidOptions.push_back(strFormatOption);
	m_ValidOptions.push_back(strGroupDefOption);
	m_ValidOptions.push_back(strScreenOption);
	m_ValidOptions.push_back(strLengthLowerBound);
	m_ValidOptions.push_back(strLengthUpperBound);
	m_ValidOptions.push_back(strForbidOrder);
	m_ValidOptions.push_back(strOrderBound);
	m_ValidOptions.push_back(strBlockOption);
	m_ValidOptions.push_back(strForbidElementOption);
	m_ValidOptions.push_back(strRequireElementOption);
	m_ValidOptions.push_back(strSpecialOrder);
}

void CommandLineOptions::AddGroup(const std::vector<int>& values)
{
	if (values.size() == 1)
		m_apGroup = std::auto_ptr<FiniteAbelianGroup>(new Zn(values[0]));
	else
	{
		std::vector<int>::const_iterator iter = values.begin();
		std::vector<int>::const_iterator iterEnd = values.end();
		std::auto_ptr<DirectSum> d(new DirectSum());
		for (; iter != iterEnd; ++iter)
			d->AddGroup(new Zn(*iter));

		m_apGroup = std::auto_ptr<FiniteAbelianGroup>(d.release());
	}

	if (m_nBlockLength == 0)
		m_nBlockLength = m_apGroup->DStar()+1;

	m_apGroup->Initialize(*this);
}

std::string CommandLineOptions::CreateDefaultFileName(eFileType type) const
{
	std::string str;
	switch (type)
	{
		case InputFile:
			return "";
		case SequenceFile:
			str = m_apGroup->GetGroupString();
			break;
		case StatsFile:
			str = m_apGroup->GetGroupString();
			str += "{Stats}";
	}
	str += ".txt";
	return str;
}

void CommandLineOptions::Parse(const OptionsList_t& rOptionsList)
{
	if (m_ValidOptions.empty())
		Initialize();

	OptionsList_t::const_iterator iter = rOptionsList.begin();
	OptionsList_t::const_iterator iterEnd = rOptionsList.end();

	str_starts_with check_str_beg;

	while (iter != iterEnd)
	{
		std::string strOption = *iter;

		iter = std::find_first_of(iter, rOptionsList.end(), m_ValidOptions.begin(), m_ValidOptions.end(), str_starts_with());
		if (iter == iterEnd)
			break;

		//Do not change the order of these!
		if (check_str_beg(strOption, strForbidOrder))
			iter = ParseForbidOrderOption(iter, iterEnd);

		else if (check_str_beg(strOption, strOrderBound))
			iter = ParseOrderBoundOption(iter, iterEnd);

		else if (check_str_beg(strOption, strSpecialOrder))
			iter = ParseSpecialOrderOption(iter, iterEnd);

		else if (check_str_beg(strOption, strForbidElementOption))
			iter = ParseForbidElementOption(iter, iterEnd);

		else if (check_str_beg(strOption, strRequireElementOption))
			iter = ParseRequireElementOption(iter, iterEnd);

		else if (check_str_beg(strOption, strInputFileOption))
			iter = ParseInputFileOption(iter, iterEnd);

		else if (check_str_beg(strOption, strStatsFileOption))
			iter = ParseStatsFileOption(iter, iterEnd);

		else if (check_str_beg(strOption, strSequenceFileOption))
			iter = ParseSequenceFileOption(iter, iterEnd);

		else if (check_str_beg(strOption, strCrossNumberOption))
			iter = ParseCrossNumberOption(iter, iterEnd);

		else if (check_str_beg(strOption, strOrderOption))
			iter = ParseOrderOption(iter, iterEnd);

		else if (check_str_beg(strOption, strMultiplicityOption))
			iter = ParseMultiplicityOption(iter, iterEnd);

		else if (check_str_beg(strOption, strFormatOption))
			iter = ParseSequenceFormatOption(iter, iterEnd);

		else if (check_str_beg(strOption, strGroupDefOption))
			iter = ParseGroupDefOption(iter, iterEnd);

		else if (check_str_beg(strOption, strScreenOption))
			iter = ParseScreenDisplayOption(iter, iterEnd);

		else if (check_str_beg(strOption, strLengthLowerBound))
			iter = ParseLowerBoundOption(iter, iterEnd);

		else if (check_str_beg(strOption, strLengthUpperBound))
			iter = ParseUpperBoundOption(iter, iterEnd);

		else if (check_str_beg(strOption, strBlockOption))
			iter = ParseBlockOption(iter, iterEnd);
	}
}

CommandLineOptions::OptionsList_t::const_iterator CommandLineOptions::ParseCrossNumberOption(OptionsList_t::const_iterator iterBegin, OptionsList_t::const_iterator iterEnd)
{
	if (!m_apGroup.get())
		throw CommandLineException("Error in Cross Number Option!  No group specified yet!", *this);

	m_bCrossNumbers = true;
	return iterBegin + 1;
}

CommandLineOptions::OptionsList_t::const_iterator CommandLineOptions::ParseScreenDisplayOption(OptionsList_t::const_iterator iterBegin, OptionsList_t::const_iterator iterEnd)
{
	if (!m_apGroup.get())
		throw CommandLineException("Error in Screen Display Option!  No group specified yet!", *this);

	std::string str = *iterBegin;

	if (str.length() > strScreenOption.length())
		m_nProgressChangeElement = ExtractNonDelimetedInteger(str, strScreenOption);

	if (m_nProgressChangeElement < 0)
		m_nProgressChangeElement = 0;

	m_bDisplayScreen = true;
	return iterBegin + 1;
}

CommandLineOptions::OptionsList_t::const_iterator CommandLineOptions::ParseOrderOption(OptionsList_t::const_iterator iterBegin, OptionsList_t::const_iterator iterEnd)
{
	m_bOrders = true;
	return iterBegin + 1;
}

CommandLineOptions::OptionsList_t::const_iterator CommandLineOptions::ParseOrderBoundOption(OptionsList_t::const_iterator iterBegin, OptionsList_t::const_iterator iterEnd)
{
	if (!m_apGroup.get())
		throw CommandLineException("Error in Order Bound Option!  No group specified yet!", *this);

	std::pair<std::string, OptionsList_t::const_iterator> pair;
	pair = ExtractDelimetedString(iterBegin, iterEnd, strOrderBound);
	std::string strBoundString = pair.first;

	std::vector<int> values;
	StringTokenizer tok;
	tok.Begin(strBoundString, ',');
	while (!tok.Done())
	{
		std::string strNum = tok.Next();
		int n;
		std::stringstream stream;
		stream << strNum;
		stream >> n;
		values.push_back(n);
	}

	if (values.size() != 2)
	{
		std::stringstream str;
		str << "Error parsing " << strOrderBound << " option.  Invalid parameters to argument";
		throw CommandLineException(str.str(), *this);
	}

	int nOrder = values[0];
	int nBound = values[1];
	m_OrderBoundLookup[nOrder] = nBound;

	return pair.second;
}

int CommandLineOptions::ExtractNonDelimetedInteger(std::string str, const std::string& strOption)
{
	int n;
	str = str.substr(strOption.length());
	std::stringstream stream;

	stream << str;
	stream >> n;

	return n;
}

CommandLineOptions::OptionsList_t::const_iterator CommandLineOptions::ParseBlockOption(OptionsList_t::const_iterator iterBegin, OptionsList_t::const_iterator iterEnd)
{
	if (!m_apGroup.get())
		throw CommandLineException("Error in Block Algorithm Option!  No group specified yet!", *this);

	std::string str = *iterBegin;

	if (str.length() > strBlockOption.length())
		m_nBlockLength = ExtractNonDelimetedInteger(str, strBlockOption);

	m_bBlockSearch = true;
	return iterBegin + 1;
}

CommandLineOptions::OptionsList_t::const_iterator CommandLineOptions::ParseForbidElementOption(OptionsList_t::const_iterator iterBegin, OptionsList_t::const_iterator iterEnd)
{
	if (!m_apGroup.get())
		throw CommandLineException("Error in Forbid Element Option!  No group specified yet!", *this);

	std::pair<std::string, OptionsList_t::const_iterator> pair;
	pair = ExtractDelimetedString(iterBegin, iterEnd, strForbidElementOption);
	std::string strElement = pair.first;

	std::vector<int>	values;
	StringTokenizer tok;
	tok.Begin(strElement, ',');
	while (!tok.Done())
	{
		int i;
		std::stringstream stream;
		std::string str = tok.Next();
		stream << str;
		stream >> i;
		values.push_back(i);
	}


	GroupElement* pElement = GetElementFromValues(values);

	m_apGroup->ForbidElement(*pElement);

	return pair.second;
}

CommandLineOptions::OptionsList_t::const_iterator CommandLineOptions::ParseSpecialOrderOption(OptionsList_t::const_iterator iterBegin, OptionsList_t::const_iterator iterEnd)
{
	if (!m_apGroup.get())
		throw CommandLineException("Error in Require Element Option!  No group specified yet!", *this);

	std::pair<std::string, OptionsList_t::const_iterator> pair;
	pair = ExtractDelimetedString(iterBegin, iterEnd, strRequireElementOption);
	std::string strElement = pair.first;

	std::vector<int>	values;
	StringTokenizer tok;
	tok.Begin(strElement, ',');
	while (!tok.Done())
	{
		int i;
		std::stringstream stream;
		std::string str = tok.Next();
		stream << str;
		stream >> i;
		values.push_back(i);
	}

	GroupElement* pElement = GetElementFromValues(values);

	pElement->SetSpecialOrdering();

	return pair.second;
}

GroupElement* CommandLineOptions::GetElementFromValues(const std::vector<int>& rValues)
{
	std::vector<int>	groupValues;
	m_apGroup->GetGroupValues(groupValues);

	if (rValues.size() != groupValues.size())
		throw CommandLineException("Mismatch in required element and group size!", *this);

	if (rValues.size() == 1)
	{
		IntModElement e(m_apGroup->order(), rValues[0]);
		return m_apGroup->GetElement(e);
	}
	else
	{
		std::vector<int>::const_iterator valueIter = rValues.begin();
		std::vector<int>::const_iterator valueIterEnd = rValues.end();
		std::vector<int>::const_iterator groupValueIter = groupValues.begin();

		DirectSumElement d(rValues.size());

		for (; valueIter != valueIterEnd; ++valueIter, ++groupValueIter)
		{
			int nValue = *valueIter;
			int nGroupValue = *groupValueIter;

			IntModElement* pNewElement = new IntModElement(nGroupValue, nValue);
			d.SetElement(valueIter - rValues.begin(), pNewElement);
		}

		return m_apGroup->GetElement(d);
	}
}

CommandLineOptions::OptionsList_t::const_iterator CommandLineOptions::ParseRequireElementOption(OptionsList_t::const_iterator iterBegin, OptionsList_t::const_iterator iterEnd)
{
	if (!m_apGroup.get())
		throw CommandLineException("Error in Require Element Option!  No group specified yet!", *this);

	std::pair<std::string, OptionsList_t::const_iterator> pair;
	pair = ExtractDelimetedString(iterBegin, iterEnd, strRequireElementOption);
	std::string strElement = pair.first;

	std::vector<int>	values;
	StringTokenizer tok;
	tok.Begin(strElement, ',');
	while (!tok.Done())
	{
		int i;
		std::stringstream stream;
		std::string str = tok.Next();
		stream << str;
		stream >> i;
		values.push_back(i);
	}

	GroupElement* pElement = GetElementFromValues(values);

	m_RequiredElements.insert(pElement);

	return pair.second;
}


CommandLineOptions::OptionsList_t::const_iterator CommandLineOptions::ParseForbidOrderOption(OptionsList_t::const_iterator iterBegin, OptionsList_t::const_iterator iterEnd)
{
	if (!m_apGroup.get())
		throw CommandLineException("Error in Forbid Order Option!  No group specified yet!", *this);

	std::string strOption = *iterBegin;
	int nDist = strOption.length() - strForbidOrder.length();
	std::string strTrunc(*iterBegin, strForbidOrder.length(), nDist);
	if (strTrunc.empty())
	{
		std::stringstream str;
		str << "Error parsing " << strForbidOrder << " option.  No order specified.";
		throw CommandLineException(str.str(), *this);
	}

	int n = ExtractNonDelimetedInteger(strTrunc, strForbidOrder);

	m_ForbiddenOrders.push_back(n);

	return iterBegin + 1;
}


CommandLineOptions::OptionsList_t::const_iterator CommandLineOptions::ParseLowerBoundOption(OptionsList_t::const_iterator iterBegin, OptionsList_t::const_iterator iterEnd)
{
	if (!m_apGroup.get())
		throw CommandLineException("Error in Lower Bound Option!  No group specified yet!", *this);

	std::string strOption = *iterBegin;
	int nDist = strOption.length() - strLengthLowerBound.length();
	std::string strTrunc(*iterBegin, strLengthLowerBound.length(), nDist);
	if (strTrunc.empty())
	{
		std::stringstream str;
		str << "Error parsing " << strLengthLowerBound << " option.  No lower bound specified.";
		throw CommandLineException(str.str(), *this);
	}

	m_nLengthLowerBound = ExtractNonDelimetedInteger(strOption, strLengthLowerBound);

	return iterBegin + 1;
}

CommandLineOptions::OptionsList_t::const_iterator CommandLineOptions::ParseUpperBoundOption(OptionsList_t::const_iterator iterBegin, OptionsList_t::const_iterator iterEnd)
{
	if (!m_apGroup.get())
		throw CommandLineException("Error in Upper Bound Option!  No group specified yet!", *this);

	std::string strOption = *iterBegin;
	int nDist = strOption.length() - strLengthUpperBound.length();
	std::string strTrunc(*iterBegin, strLengthUpperBound.length(), nDist);
	if (strTrunc.empty())
	{
		std::stringstream str;
		str << "Error parsing " << strLengthUpperBound << " option.  No upper bound specified.";
		throw CommandLineException(str.str(), *this);
	}

	m_nLengthUpperBound = ExtractNonDelimetedInteger(strOption, strLengthUpperBound);

	return iterBegin + 1;
}

CommandLineOptions::OptionsList_t::const_iterator CommandLineOptions::ParseMultiplicityOption(OptionsList_t::const_iterator iterBegin, OptionsList_t::const_iterator iterEnd)
{
	if (!m_apGroup.get())
		throw CommandLineException("Error in Multiplicity Option!  No group specified yet!", *this);

	m_bMultiplicity = true;
	return iterBegin + 1;
}

CommandLineOptions::OptionsList_t::const_iterator CommandLineOptions::ParseSequenceFormatOption(OptionsList_t::const_iterator iterBegin, OptionsList_t::const_iterator iterEnd)
{
	if (!m_apGroup.get())
		throw CommandLineException("Error in Sequence Format Option!  No group specified yet!", *this);

	std::string strTrunc(*iterBegin, strFormatOption.length(), 1);
	if (strTrunc.empty())
	{
		std::stringstream str;
		str << "Error parsing " << strFormatOption << " option.  No format option specified.";
		throw CommandLineException(str.str(), *this);
	}

	std::stringstream stream(strTrunc);
	int n;
	stream >> n;
	if (n != DispFormatStandard && n != DispFormatMultiplicity)
	{
		std::stringstream str;
		str << "Error parsing " << strFormatOption << " option.  Invalid format option specified.";
		throw CommandLineException(str.str(), *this);
	}

	m_eDisplayFormat = static_cast<eDispFormat>(n);
	return iterBegin + 1;
}

CommandLineOptions::OptionsList_t::const_iterator CommandLineOptions::ParseInputFileOption(OptionsList_t::const_iterator iter, OptionsList_t::const_iterator iterEnd)
{
	if (!m_apGroup.get())
		throw CommandLineException("Error in Input File Option!  No group specified yet!", *this);

	std::pair<std::string, OptionsList_t::const_iterator> pair;
	pair = ExtractDelimetedString(iter, iterEnd, strInputFileOption);
	m_strInFile = pair.first;

	return pair.second;
}

CommandLineOptions::OptionsList_t::const_iterator CommandLineOptions::ParseSequenceFileOption(OptionsList_t::const_iterator iter, OptionsList_t::const_iterator iterEnd)
{
	if (!m_apGroup.get())
		throw CommandLineException("Error in Sequence File Option!  No group specified yet!", *this);

	
	std::string strFirst = *iter;
	strFirst = strFirst.substr(strSequenceFileOption.length());
	if (strFirst.length() == 1 && (strFirst == "X" || strFirst == "x"))
	{
		m_strSequenceFile = "";
		m_bUseSequenceFile = false;
		return iter + 1;
	}

	std::pair<std::string, OptionsList_t::const_iterator> pair;
	pair = ExtractDelimetedString(iter, iterEnd, strSequenceFileOption);
	m_strSequenceFile = pair.first;

	return pair.second;
}

CommandLineOptions::OptionsList_t::const_iterator CommandLineOptions::ParseStatsFileOption(OptionsList_t::const_iterator iter, OptionsList_t::const_iterator iterEnd)
{
	if (!m_apGroup.get())
		throw CommandLineException("Error in Stats File Option!  No group specified yet!", *this);

	std::pair<std::string, OptionsList_t::const_iterator> pair;
	pair = ExtractDelimetedString(iter, iterEnd, strStatsFileOption);
	m_strStatsFile = pair.first;

	return pair.second;
}

std::pair<std::string, CommandLineOptions::OptionsList_t::const_iterator> CommandLineOptions::ExtractDelimetedString(OptionsList_t::const_iterator iter, OptionsList_t::const_iterator iterEnd, const std::string& strOptionCmd, char d1, char d2)
{
	std::string str;

	std::string strFirst = *iter;
	std::string::size_type pos = strFirst.find('[', 0);
	if (pos != strOptionCmd.length())
	{
		std::stringstream str;
		str << "Error parsing " << strFormatOption << " option.  [] delimeters not found.";
		throw CommandLineException(str.str(), *this);
	}

	OptionsList_t::const_iterator iterClose;

	std::string::size_type pos2 = strFirst.find(']', pos+1);

	std::stringstream streamFileName;
	if (pos2 != std::string::npos)
		iterClose = iter+1;
	else
		iterClose = std::find_if(iter+1, iterEnd, std::bind2nd(str_contains<char>(), ']'))+1;

	while (iter != iterClose)
	{
		streamFileName << *iter;

		if (++iter != iterClose)
			streamFileName << " ";
	}

	str = streamFileName.str();
	pos = str.find('[');
	pos2 = str.find(']', pos+1);

	if (pos == std::string::npos || pos2 == std::string::npos)
	{
		std::stringstream str;
		str << "Error parsing " << strFormatOption << " option.  [] delimeters not found.";
		throw CommandLineException(str.str(), *this);
	}

	str = str.substr(pos+1, pos2-pos-1);

	return std::make_pair(str, iterClose);
}

CommandLineOptions::OptionsList_t::const_iterator CommandLineOptions::ParseGroupDefOption(OptionsList_t::const_iterator iter, OptionsList_t::const_iterator iterEnd)
{
	std::pair<std::string, OptionsList_t::const_iterator> pair;

	pair = ExtractDelimetedString(iter, iterEnd, strGroupDefOption);

	std::string strGroupString = pair.first;

	std::vector<int>	values;

	StringTokenizer tok;

	tok.Begin(strGroupString, ' ');
	while (!tok.Done())
	{
		std::string strTok = tok.Next();
		std::stringstream stream;
		stream << strTok;
		int i;
		stream >> i;
		values.push_back(i);
	}

	std::vector<int>::const_iterator iter2 = values.begin();
	std::vector<int>::const_iterator iter2End = values.end();
	while (iter2 != iter2End)
	{
		int nCur = *iter2;

		++iter2;
		if (iter2 == iter2End)
			break;

		int nNext = *iter2;

		++iter2;

		if (nNext % nCur != 0)
		{
			std::stringstream strError;
			strError << "Error with group specification [" << strGroupString << "].  Program requires m{n} | m{n+1}.";
			throw CommandLineException(strError.str(), *this);
		}
	}

	AddGroup(values);

	return pair.second;
}

std::string CommandLineOptions::GetInputFile() const
{
	if (!m_strInFile.empty())
		return m_strInFile;

	return CreateDefaultFileName(InputFile);
}

std::string CommandLineOptions::GetSequenceFile() const
{
	if (!m_strSequenceFile.empty())
		return m_strSequenceFile;

	return CreateDefaultFileName(SequenceFile);
}

std::string CommandLineOptions::GetStatsFile() const
{
	if (!m_strStatsFile.empty())
		return m_strStatsFile;

	return CreateDefaultFileName(StatsFile);
}


bool CommandLineOptions::UseInputFile() const
{
	return (!GetInputFile().empty());
}

bool CommandLineOptions::UseSequenceFile() const
{
	return m_bUseSequenceFile;
}

bool CommandLineOptions::UseStatsFile() const
{
	return (!GetStatsFile().empty());
}


bool CommandLineOptions::IsOrderBounded(int order) const
{
	OrderBoundLookup_t::const_iterator iter = m_OrderBoundLookup.find(order);
	return (iter != m_OrderBoundLookup.end());
}

int CommandLineOptions::GetOrderBound(int nOrder) const
{
	OrderBoundLookup_t::const_iterator iter = m_OrderBoundLookup.find(nOrder);
	if (iter == m_OrderBoundLookup.end())
		throw class CommandLineException("Error", *this);

	return iter->second;
}


std::string CommandLineOptions::GetHelpString() const
{
	std::stringstream stream;
	stream << "Zero Sequencer	2003 Trinity University REU (Programmed by Zachary Turner)" << std::endl;
	stream << std::endl;
	stream << "Note: Commands are case sensitive." << std::endl;
	stream << "Note: " << strGroupDefOption << " must come before any other options." << std::endl;
	stream << std::endl;
	stream << "Command Line Switches " << std::endl;
	stream << "\t" << strSequenceFileOption << "[file.txt]\tOutputs sequence information to \"file.txt\"" << std::endl;
	stream << "\t" << strSequenceFileOption << "X\t\tDon't output sequence data at all" << std::endl;
	stream << "\t" << strStatsFileOption << "[file.txt]\tOutputs statistics information to \"file.txt\"" << std::endl;
	stream << "\t" << strOrderOption << "\t\tCalculates orders class statistics for the group" << std::endl;
	stream << "\t" << strMultiplicityOption << "\t\tCalculates multiplicity class statistics for the group" << std::endl;
	stream << "\t" << strFormatOption << DispFormatStandard << "\t\tDisplay atoms in standard format" << std::endl;
	stream << "\t" << strFormatOption << DispFormatMultiplicity << "\t\tDisplay atoms in run length encoded format" << std::endl;
	stream << "\t" << strGroupDefOption << "[m1 ... mn]\tSpecify group, (i.e. Z_m1+...+Z_mn)" << std::endl;
	stream << "\t" << strScreenOption << "\t\tDisplay output to screen" << std::endl;
	stream << "\t" << strScreenOption << "n\t\tDisplay progress whenever the k <= n'th element changes" << std::endl;
	stream << "\t" << strLengthLowerBound << "n\t\tOnly search for sequences of length >= n" << std::endl;
	stream << "\t" << strForbidOrder << "n\t\tForbid any element E where order(E) = n" << std::endl;
	stream << "\t" << strOrderBound << "[n,m]\tForbids any order n elements from appearing k>=m times" << std::endl;
	stream << "\t" << strRequireElementOption << "[e]\t\tForces 'e' to appear in the sequence at least once" << std::endl;
	stream << "\t" << strForbidElementOption << "[e]\t\tForbids 'e' from appearing in the sequence" << std::endl;

	return stream.str();
}
