#include "stdafx.h"
#include "GroupDef.h"
#include "SequenceElement.h"
#include "Utility.h"
#include "SequenceClasses.h"
#include "CommandLine.h"

Group::Group()
	: m_nExponent(0)
{
}

GroupElement* Group::InverseCombine(const GroupElement& lhs, const GroupElement& rhs) const
{
	return CombineElements(lhs, *Inverse(rhs));
}

void Group::EncodeSequence(const Sequence_t& s, CodedSequence_t& cs) const
{
	Multiset_t set(s.begin(), s.end());
	Multiset_t::const_iterator iter = set.begin();
	Multiset_t::const_iterator iterEnd = set.end();
	while (iter != iterEnd)
	{
		const Multiset_t::value_type& rVal = *iter;
		int n = set.count(rVal);
		cs.insert(std::make_pair(rVal, n));
		std::advance(iter, n);
	}
}

void Group::SetExponent(int nExp)
{
	m_nExponent = nExp;

	m_nDivisorList.clear();

	int bound = ceil(sqrt((double)nExp));

	m_nDivisorList.push_back(1);
	for (int n = 2; n <= bound; ++n)
	{
		if (nExp % n == 0)
		{
			m_nDivisorList.push_back(n);
			m_nDivisorList.push_back(nExp / n);
		}
	}
	m_nDivisorList.push_back(nExp);
}



OrderClass* Group::GetOrderClass(const Sequence_t& rElement)
{
	Sequence_t::const_iterator iter = rElement.begin();
	Sequence_t::const_iterator iterEnd = rElement.end();

	//Maybe we can do lazy (or pre) evaluation on element orders
	OrderClass o(m_nDivisorList);
	for (; iter != iterEnd; ++iter)
	{
		GroupElement& rElement = **iter;

		int nOrder = order(rElement);

		o[nOrder]++;
	}

	
	OrderClass* pRet = NULL;

	Group* pGroup = const_cast<Group*>(this);

	OrderClassSet_t::const_iterator ocIter = m_OrderClassSet.find(&o);
	if (ocIter == m_OrderClassSet.end())
		pGroup->m_OrderClassSet.insert(pRet = new OrderClass(o));
	else
		pRet = *ocIter;

	return pRet;
}

MultiplicityClass* Group::GetMultiplicityClass(const CodedSequence_t& rElement)
{

	MultiplicityClass m(rElement);

	
	MultiplicityClass* pRet = NULL;

	Group* pGroup = const_cast<Group*>(this);

	MultiplicityClassSet_t::const_iterator mcIter = m_MultiplicityClassSet.find(&m);
	if (mcIter == m_MultiplicityClassSet.end())
		pGroup->m_MultiplicityClassSet.insert(pRet = new MultiplicityClass(m));
	else
		pRet = *mcIter;

	return pRet;
}





FiniteAbelianGroup::~FiniteAbelianGroup()
{
	iterator iter = m_Group.begin();
	iterator iterEnd = m_Group.end();

	for (; iter != iterEnd; ++iter)
	{
		std::auto_ptr<GroupElement> ap(*iter);
	}

	m_Group.clear();
	m_SequencePool.clear();
}

int FiniteAbelianGroup::order(GroupElement& rElem)
{
	ElmtOrderLookup_t::const_iterator iter = m_OrderLookup.find(&rElem);
	if (iter == m_OrderLookup.end())
		return (m_OrderLookup[&rElem] = order_impl(rElem, rElem));

	return (*iter).second;
}

int FiniteAbelianGroup::order_impl(GroupElement& rElem, const GroupElement& rBasis)
{
	GroupElement* pCombined = CombineElements(rElem, rBasis);

	GroupElement* rZero = GetZero();
	if (rElem == *rZero)
		return 1;

	return 1 + order_impl(*pCombined, rBasis);
}


int	FiniteAbelianGroup::ElementToPosition(const GroupElement& rElement) const
{
	GroupElement* p = const_cast<GroupElement*>(&rElement);
	Set_t::const_iterator iter = m_Group.find(p);
	return std::distance(m_Group.begin(), iter);
}

GroupElement* FiniteAbelianGroup::PositionToElement(int nPosition) const
{
	Set_t::const_iterator iter = m_Group.begin();

	std::advance(iter, nPosition);
	return *iter;
}

void FiniteAbelianGroup::ForbidElement(GroupElement& rElement)
{
	m_SequencePool.erase(&rElement);
}



GroupElement* FiniteAbelianGroup::GetElement(const GroupElement& rElement) const
{
	GroupElement* p = const_cast<GroupElement*>(&rElement);
	
	Set_t::const_iterator iter = m_Group.find(p);

	if (iter == m_Group.end())
		throw;

	return *iter;
}


void Zn::Initialize(const CommandLineOptions& o)
{
	m_Group.insert(new IntModElement(m_n, 0));
	for (int i=1; i < m_n; ++i)
	{
		m_Group.insert(new IntModElement(m_n, i));
	}

	SetExponent(m_n);

	Set_t::const_iterator iter = m_Group.begin();
	Set_t::const_iterator iterEnd = m_Group.end();
	for (; iter != iterEnd; ++iter)
	{
		GroupElement& rElem = **iter;
		int nOrder = order(rElem);
		if (nOrder == 1)
			continue;
		CommandLineOptions::ForbiddenOrderIter_t iter2 = std::find(o.forbidden_order_begin(), o.forbidden_order_end(), nOrder);

		if (iter2 == o.forbidden_order_end())
			m_SequencePool.insert(&rElem);
	}
}

std::string Zn::GetGroupString() const
{
	std::stringstream str;
	str << m_n;
	return str.str();
}


void Zn::GetGroupValues(std::vector<int>& rValues) const
{
	rValues.push_back(m_n);
}

GroupElement* Zn::Inverse(const GroupElement& rGroupElement) const
{
	const IntModElement& rElement = debug_dynamic_cast<const IntModElement&>(rGroupElement);

	return GetElement(IntModElement(m_n, m_n - rElement.GetVal()));
}

GroupElement* Zn::CombineElements(const GroupElement& lhs, const GroupElement& rhs) const
{
	const IntModElement& lhs1 = debug_dynamic_cast<const IntModElement&>(lhs);
	const IntModElement& rhs1 = debug_dynamic_cast<const IntModElement&>(rhs);

	return GetElement(lhs1 + rhs1);
}


GroupElement* Zn::GetZero() const
{
	return GetElement(IntModElement(m_n, 0));
}

int	Zn::DStar() const
{
	return m_n - 1;
}

bool DirectSum::IncrementIteratorList(IteratorList_t& rList)
{
	IteratorList_t::reverse_iterator listIter = rList.rbegin();
	IteratorList_t::reverse_iterator listIterEnd = rList.rend();

	for (;listIter != listIterEnd; ++listIter)
	{
		IterPair_t& rCurPair = *listIter;


		GroupList_t::iterator& rGroupIter = rCurPair.first;
		FiniteAbelianGroup::iterator& elmtIter = rCurPair.second;

		++elmtIter;
		if (elmtIter != (*rGroupIter)->end())
			return false;

		elmtIter = (*rGroupIter)->begin();
	}

	return true;
}


void DirectSum::Initialize(const CommandLineOptions& o)
{
	IteratorList_t	iteratorList;

	{
		GroupList_t::iterator iter = m_pGroupList.begin();
		GroupList_t::iterator iterEnd = m_pGroupList.end();

		for (; iter != iterEnd; ++iter)
		{
			(*iter)->Initialize(o);
			iteratorList.push_back(std::make_pair(iter, (*iter)->begin()));
		}
	}

	//First Generate the zero element
	DirectSumElement d(m_pGroupList.size());

	GroupList_t::const_iterator iter = m_pGroupList.begin();
	GroupList_t::const_iterator iterEnd = m_pGroupList.end();
	for (; iter != iterEnd; ++iter)
	{
		GroupElement* pZero = (*iter)->GetZero();
		d.SetElement(iter - m_pGroupList.begin(), pZero);
	}
	DirectSumElement* pZero = new DirectSumElement(d);
	m_Group.insert(pZero);


	bool bFinished = false;
	while (!bFinished)
	{
		//Make the DirectSumElement that the current sequence of iterators represents
		IteratorList_t::const_iterator iter = iteratorList.begin();
		IteratorList_t::const_iterator iterEnd = iteratorList.end();
		DirectSumElement d(iteratorList.size());

		for (; iter != iterEnd; ++iter)
		{
			int nIndex = iter - iteratorList.begin();
			GroupElement* pElement = *(*iter).second;

			d.SetElement(nIndex, pElement);
		}


		bFinished = IncrementIteratorList(iteratorList);
		if (d == *pZero)
			continue;

		DirectSumElement* pD = new DirectSumElement(d);
		m_Group.insert(pD);
	}

	//Calculate the exponent of the group
	iter = m_pGroupList.begin();
	iterEnd = m_pGroupList.end();

	int nExp = 0;
	for (; iter != iterEnd; ++iter)
		nExp = _max((*iter)->GetExponent(), nExp);
	SetExponent(nExp);


	Set_t::const_iterator groupIter = m_Group.begin();
	Set_t::const_iterator groupIterEnd = m_Group.end();
	for (; groupIter != groupIterEnd; ++groupIter)
	{
		GroupElement& rElem = **groupIter;
		int nOrder = order(rElem);
		if (nOrder == 1)
			continue;
		CommandLineOptions::ForbiddenOrderIter_t iter2 = std::find(o.forbidden_order_begin(), o.forbidden_order_end(), nOrder);

		if (iter2 == o.forbidden_order_end())
			m_SequencePool.insert(&rElem);
	}
}

std::string DirectSum::GetGroupString() const
{
	GroupList_t::const_iterator iter = m_pGroupList.begin();
	GroupList_t::const_iterator iterEnd = m_pGroupList.end();

	std::stringstream str;
	while (iter != iterEnd)
	{
		const FiniteAbelianGroup& rElement = **iter;
		str << rElement.GetGroupString();

		if (++iter != iterEnd)
			str << " ";
	}

	return str.str();
}

void DirectSum::GetGroupValues(std::vector<int>& rValues) const
{
	GroupList_t::const_iterator iter = m_pGroupList.begin();
	GroupList_t::const_iterator iterEnd = m_pGroupList.end();

	std::stringstream str;
	while (iter != iterEnd)
	{
		std::vector<int> tempValues;
		const FiniteAbelianGroup& rElement = **iter;
		rElement.GetGroupValues(tempValues);

		//std::cout << "Before copy.  LHS = ";
		//std::copy(rValues.begin(), rValues.end(), std::ostream_iterator<int>(std::cout, " "));
		//std::cout << ".  RHS = ";
		//std::copy(tempValues.begin(), tempValues.end(), std::ostream_iterator<int>(std::cout, " "));

		int nIndex = rValues.size();
		rValues.resize(rValues.size() + tempValues.size());

		std::copy(tempValues.begin(), tempValues.end(), &rValues[nIndex]);

		//std::cout << "After copy.  LHS = ";
		//std::copy(rValues.begin(), rValues.end(), std::ostream_iterator<int>(std::cout, " "));

		++iter;
	}
}

GroupElement* DirectSum::Inverse(const GroupElement& rGroupElement) const
{
	const DirectSumElement& rElement = debug_dynamic_cast<const DirectSumElement&>(rGroupElement);

	DirectSumElement::const_iterator iter = rElement.begin();
	DirectSumElement::const_iterator iterEnd = rElement.end();

	DirectSumElement inverse(rElement.length());

	GroupList_t::const_iterator groupIter = m_pGroupList.begin();
	for (; iter != iterEnd; ++iter, ++groupIter)
	{
		GroupElement* pInverse = (*groupIter)->Inverse(**iter);
		inverse.SetElement(iter - rElement.begin(), pInverse);
	}

	return GetElement(inverse);
}

GroupElement* DirectSum::GetZero() const
{
	DirectSumElement d(m_pGroupList.size());

	GroupList_t::const_iterator iter = m_pGroupList.begin();
	GroupList_t::const_iterator iterEnd = m_pGroupList.end();
	for (; iter != iterEnd; ++iter)
	{
		GroupElement* pZero = (*iter)->GetZero();
		d.SetElement(iter - m_pGroupList.begin(), pZero);
	}

	return GetElement(d);
}

int	DirectSum::DStar() const
{
	GroupList_t::const_iterator iter = m_pGroupList.begin();
	GroupList_t::const_iterator iterEnd = m_pGroupList.end();
	int dstar = 0;
	for (; iter != iterEnd; ++iter)
	{
		dstar += (*iter)->DStar();
	}

	return dstar + 1;
}


GroupElement* DirectSum::CombineElements(const GroupElement& lhs, const GroupElement& rhs) const
{
	const DirectSumElement& lhs1 = debug_dynamic_cast<const DirectSumElement&>(lhs);
	const DirectSumElement& rhs1 = debug_dynamic_cast<const DirectSumElement&>(rhs);

	DirectSumElement::const_iterator iter = lhs1.begin();
	DirectSumElement::const_iterator iterBegin = lhs1.begin();
	DirectSumElement::const_iterator iterEnd = lhs1.end();
	DirectSumElement::const_iterator iter2 = rhs1.begin();
	DirectSumElement::const_iterator iter2End = rhs1.end();

	DirectSumElement sum(lhs1.length());

	GroupList_t::const_iterator groupIter = m_pGroupList.begin();
	for (; iter != iterEnd; ++iter, ++iter2, ++groupIter)
	{
		sum.SetElement(iter - iterBegin, (*groupIter)->CombineElements(**iter, **iter2));
	}

	return GetElement(sum);
}

