#ifndef __GROUPDEF_H__
#define __GROUPDEF_H__

#include "algorithm.h"
#include "TypeDef.h"

class GroupElement;

class OrderClass;

class MultiplicityClass;
class CommandLineOptions;

class Group
{
public:
	Group();

	virtual ~Group() {}

	virtual void Initialize(const CommandLineOptions& o) = 0;
	
	typedef Set_t::const_iterator		const_iterator;
	typedef Set_t::iterator				iterator;
	typedef Set_t::size_type			size_type;

	virtual std::string GetGroupString() const = 0;
	virtual void GetGroupValues(std::vector<int>& rValues) const = 0;

	virtual const_iterator begin()	const	= 0;
	virtual const_iterator end()	const	= 0;
	virtual iterator begin()				= 0;
	virtual iterator end()					= 0;
	virtual int order()		const	= 0;
	virtual int order(GroupElement& rElem)  = 0;

	virtual GroupElement*	Inverse(const GroupElement& rGroupElement) const = 0;
	virtual GroupElement*	CombineElements(const GroupElement& lhs, const GroupElement& rhs) const = 0;
	virtual GroupElement*	InverseCombine(const GroupElement& lhs, const GroupElement& rhs) const;

	virtual GroupElement*	GetZero() const = 0;
	virtual int				DStar() const = 0;

	OrderClass*	GetOrderClass(const Sequence_t& rElement);
	MultiplicityClass* GetMultiplicityClass(const CodedSequence_t& rElement);


	void EncodeSequence(const Sequence_t& s, CodedSequence_t& cs) const;

	int GetExponent() const	{return m_nExponent;}

protected:
	void				SetExponent(int nExp);
	std::vector<int>	m_nDivisorList;

private:
	typedef std::set<OrderClass*, val_less<OrderClass*> >	OrderClassSet_t;
	typedef std::set<MultiplicityClass*, val_less<MultiplicityClass*> >	MultiplicityClassSet_t;

	OrderClassSet_t			m_OrderClassSet;
	MultiplicityClassSet_t	m_MultiplicityClassSet;
	int						m_nExponent;
};

class FiniteAbelianGroup : 
	public Group
{
public:	
	FiniteAbelianGroup(const Set_t& rSet)
		: m_Group(rSet)
	{
	}

	FiniteAbelianGroup() {}

	virtual ~FiniteAbelianGroup();

	virtual const_iterator begin()	const	{return m_Group.begin();}
	virtual const_iterator end()	const	{return m_Group.end();}
	virtual iterator begin()				{return m_Group.begin();}
	virtual iterator end()					{return m_Group.end();}
	virtual int order()		const	{return m_Group.size();}
	virtual int order(GroupElement& rElem);

	int				ElementToPosition(const GroupElement& rElement) const;
	GroupElement*	PositionToElement(int nPosition) const;

	const_iterator pool_begin()	const	{return m_SequencePool.begin();}
	const_iterator pool_end()	const	{return m_SequencePool.end();}
	iterator pool_begin()				{return m_SequencePool.begin();}
	iterator pool_end()					{return m_SequencePool.end();}
	size_type pool_size()		const	{return m_SequencePool.size();}

	GroupElement*	GetElement(const GroupElement& rElement) const;
	void			ForbidElement(GroupElement& rElement);

protected:
	Set_t	m_Group;
	Set_t	m_SequencePool;
	ElmtOrderLookup_t	m_OrderLookup;

private:
	int order_impl(GroupElement& rElem, const GroupElement& rBasis);

};

class Zn
	: public FiniteAbelianGroup
{
public:
	Zn(int n)
		: m_n(n)
	{
	}

	virtual std::string GetGroupString() const;
	virtual void GetGroupValues(std::vector<int>& rValues) const;

	virtual void Initialize(const CommandLineOptions& o);

	virtual GroupElement*	Inverse(const GroupElement& rGroupElement) const;
	virtual GroupElement*	CombineElements(const GroupElement& lhs, const GroupElement& rhs) const;

	virtual GroupElement*	GetZero() const;
	virtual int				DStar() const;

private:
	int m_n;
};

class DirectSum 
	: public FiniteAbelianGroup
{
	typedef std::vector<FiniteAbelianGroup*>	GroupList_t;
public:
	DirectSum()
	{
	}

	~DirectSum()
	{
	}

	virtual void Initialize(const CommandLineOptions& o);

	virtual std::string GetGroupString() const;
	virtual void GetGroupValues(std::vector<int>& rValues) const;

	virtual GroupElement*	Inverse(const GroupElement& rGroupElement) const;
	virtual GroupElement*	CombineElements(const GroupElement& lhs, const GroupElement& rhs) const;

	virtual GroupElement*	GetZero() const;
	virtual int				DStar() const;

	void AddGroup(FiniteAbelianGroup* pGroup)	{m_pGroupList.push_back(pGroup);}


private:
	typedef std::pair<GroupList_t::iterator, FiniteAbelianGroup::iterator>	IterPair_t;
	typedef std::vector<IterPair_t>	IteratorList_t;

	bool IncrementIteratorList(IteratorList_t& rList);

	GroupList_t		m_pGroupList;
};


#endif